import * as d3 from 'd3'
import {Color, PI, select} from './utils'

const ARC_CLASS = 'd3d-arc'
const SLICE_CLASS = ARC_CLASS + '-slice'
const PART_CLASS = SLICE_CLASS + '-part'

/**************************************************************************************
 * arc
 **************************************************************************************/

/**
 * 展开参数的add函数
 */
export function add(parent, startAngle, endAngle, rx, ry, h, ir = .5, color = 'green', edgeCount = 5) {
  const option = {startAngle, endAngle, rx, ry, h, ir, color, edgeCount}
  return _add(parent, option)
}

/**
 * 添加一个g元素代表一个arc，并更新arc
 */
export function _add(parent, option) {
  const arc = select(parent).append('g').attr('class', ARC_CLASS)
  // .attr('transform', `translate(0, -${h / 2})`)

  update(arc, option)
  return arc
}

/**
 * 更新过程的transition封装
 */
export function update(arc, option) {
  select(arc).transition().duration(300).tween('fill', function () {
    const oldOption = this._option || {...option, startAngle: 0, endAngle: 0}
    const interpolator = d3.interpolate(oldOption, option)
    this._option = option
    return t => {
      const opt = interpolator(t)
      _update(this, opt)
    }
  })
  return select(arc)
}

/**
 * 更新一个arc，按顺序创建或更新8个面
 */
export function _update(arc, option) {
  const [angleNear, angleFar] = splitAngle(option.startAngle, option.endAngle, option.edgeCount)
  const color = option.color
  const sliceOptions = [
    [pathTop, 'bottom', {h: 0, color: Color.bottom(color)}],
    [pathSide, 'outer far', {...angleFar, ir: 1, color: Color.outerFar(color)}],
    [pathSection, 'section', {angle: option.endAngle, color: Color.section(color)}],
    [pathSection, 'section', {angle: option.startAngle, color: Color.section(color)}],
    [pathSide, 'inner far', {...angleFar, color: Color.innerFar(color)}],
    [pathSide, 'inner near', {...angleNear, color: Color.inner(color)}],
    [pathSide, 'outer near', {...angleNear, ir: 1, color: Color.outer(color)}],
    [pathTop, 'top', {color: Color.top(color)}],
  ]
  const update = select(arc).selectAll('.' + SLICE_CLASS).data(sliceOptions)
    .each(function ([pathFunc, _, sliceOpt]) {
      updateSlice(pathFunc, this, {...option, ...sliceOpt})
    })
  update.enter().append('g')
    .attr('class', ([_, clz]) => `${SLICE_CLASS} ${clz}`)
    .each(function ([pathFunc, _, sliceOpt]) {
      updateSlice(pathFunc, this, {...option, ...sliceOpt})
    })
  return select(arc)
}

/**************************************************************************************
 * slice 画弧的各个面
 **************************************************************************************/
/**
 * 添加一个g元素代表一个slice, 并更新slice
 */
export function addSlice(pathFunc, parent, option) {
  const slice = select(parent).append('g').attr('class', 'd3d-arc-slice')
  return updateSlice(pathFunc, slice, option)
}

/**
 * 画顶面的updateSlice别名函数
 */
export function updateTop(parent, option) {
  return updateSlice(pathTop, parent, option)
  // .attr('class', 'd3d-arc-slice-' + (option.h === 0 ? 'top' : 'bottom'))
}

/**
 * 画侧面的updateSlice别名函数
 */
export function updateSide(parent, option) {
  return updateSlice(pathSide, parent, option)
  // .attr('class', 'd3d-arc-slice-' + (option.ir === 1 ? 'outer' : 'inner'))
}

export function updateSlice(pathFunc, slice, option) {
  const angleParts = [{startAngle: option.startAngle, endAngle: option.endAngle}]
  // .filter(angle => angle.startAngle !== angle.endAngle)

  const update = select(slice).selectAll('.' + PART_CLASS).data(angleParts)
  update
    .attr('stroke', option.color)
    .attr('fill', option.color)
    .each(function (angle) {
      const partOption = {...option, ...angle}
      updatePart(pathFunc, this, partOption)
    })
  update.enter().append('path')
    .attr('stroke', option.color)
    .attr('fill', option.color)
    .attr('class', angle => PART_CLASS + (angle.startAngle < PI ? ' near' : ' far'))
    .each(function (angle) {
      const partOption = {...option, ...angle}
      updatePart(pathFunc, this, partOption)
    })
  update.exit().remove()
  return slice
}

function updatePart(pathFunc, part, option) {
  select(part)
    .attr('d', d => {
      // .transition().duration(1500).attrTween('d', d => {
      // part._option = part._option || {...option, startAngle: 0, endAngle: 0}
      // const interpolator = d3.interpolate(part._option, option)
      // return t => {
      //   part._option = option
      // const opt = interpolator(t)
      return pathFunc(option)
      // }
    })
  // : select(part).attr('d', d => {
  //   part._option = option
  //   return pathFunc(option)
  // })
}

/**************************************************************************************
 * path
 **************************************************************************************/

/**
 * 构建弧的顶面路径
 */
export function pathTop(option) {
  return option.edgeCount ? pathPrismTop(option) : pathArcTop(option)
}

function pathArcTop({startAngle, endAngle, rx, ry, h, ir}) {
  if (startAngle === endAngle) {
    return 'M 0 0'
  }

  if (startAngle === 0 && endAngle === 2 * PI) {
    endAngle = 0.9999999 * endAngle
  }

  const sx = rx * Math.cos(startAngle)
  const sy = ry * Math.sin(startAngle) - h
  const ex = rx * Math.cos(endAngle)
  const ey = ry * Math.sin(endAngle) - h

  const isx = ir * rx * Math.cos(startAngle)
  const isy = ir * ry * Math.sin(startAngle) - h
  const iex = ir * rx * Math.cos(endAngle)
  const iey = ir * ry * Math.sin(endAngle) - h

  return [
    'M', sx, sy,
    'A', rx, ry, '0', (endAngle - startAngle > Math.PI ? 1 : 0), '1', ex, ey, // 外弧
    'L', iex, iey, // 外弧到内弧的边
    'A', ir * rx, ir * ry, '0', (endAngle - startAngle > Math.PI ? 1 : 0), '0', isx, isy, // 内弧
    'z'
  ].join(' ')
}

function pathPrismTop({startAngle, endAngle, rx, ry, h, ir, edgeCount}) {
  const anglePerPoint = (endAngle - startAngle) / edgeCount
  let points = [...new Array(edgeCount + 1)]
    .map((_, i) => i * anglePerPoint + startAngle)
    .map(rad => ({x: rx * Math.cos(rad), y: ry * Math.sin(rad)}))

  const innerPoints = points.map(({x, y}) => ({x: x * ir, y: y * ir})).reverse()
  points = points.concat(innerPoints)
  points.forEach(p => p.y -= h)

  const path = points
    .map((p, i) => ['L', p.x, p.y])
    .reduce((total, item) => total.concat(item), [])
  path.push('z')
  path[0] = 'M'

  return path.join(' ')
}

/**
 * 构建弧的侧面路径
 */
export function pathSide(option) {
  return option.edgeCount ? pathPrismSide(option) : pathArcSide(option)
}

export function pathArcSide({startAngle, endAngle, rx, ry, h, ir = 1}) {
  if (startAngle === endAngle) {
    return 'M 0 0'
  }

  h = -h
  rx *= ir
  ry *= ir
  const [sx, sy, ex, ey] = [
    rx * Math.cos(startAngle),
    ry * Math.sin(startAngle),
    rx * Math.cos(endAngle),
    ry * Math.sin(endAngle),
  ]
  return [
    'M', sx, h + sy,
    'A', rx, ry, '0 0 1', ex, h + ey, // 下底弧
    'L', ex, ey, // 结束边
    'A', rx, ry, '0 0 0', sx, sy, // 上底弧
    'L', sx, h + sy, // 起始边
    'z'
  ].join(' ')
}

function pathPrismSide({startAngle, endAngle, rx, ry, h, ir, edgeCount}) {
  console.log('pathPrismSide', startAngle, endAngle, edgeCount)
  if (startAngle === endAngle) {
    return 'M 0 0'
  }

  const anglePerPoint = (endAngle - startAngle) / edgeCount
  let points = [...new Array(edgeCount + 1)]
    .map((_, i) => i * anglePerPoint + startAngle)
    .map(rad => ({x: ir * rx * Math.cos(rad), y: ir * ry * Math.sin(rad)}))
  const topPoints = points.map(p => ({x: p.x, y: p.y - h})).reverse()
  points = points.concat(topPoints)

  const path = points
    .map((p, i) => ['L', p.x, p.y])
    .reduce((total, item) => total.concat(item), [])
  path.push('z')
  path[0] = 'M'

  return path.join(' ')
}

/**
 * 构建弧的横截面路径
 * @param angle
 * @param rx
 * @param ry
 * @param h
 * @param ir
 * @returns {string}
 */
export function pathSection({angle, rx, ry, h, ir = 1}) {
  const [sxi, syi, exo, eyo] = [
    ir * rx * Math.cos(angle),
    ir * ry * Math.sin(angle),
    rx * Math.cos(angle),
    ry * Math.sin(angle),
  ]
  return ['M', sxi, syi,
    'L', sxi, syi - h,
    'L', exo, eyo - h,
    'L', exo, eyo,
    'z'
  ].join(' ')
}

function splitAngle(startAngle, endAngle, edgeCount) {
  if (startAngle < PI && PI < endAngle) {
    if (!edgeCount) {
      return [{startAngle, endAngle: PI}, {startAngle: PI, endAngle}]
    }

    const anglePerPoint = (endAngle - startAngle) / edgeCount
    const pointsX = [...new Array(edgeCount + 1)]
      .map((_, i) => i * anglePerPoint + startAngle)
      .map(rad => Math.cos(rad))
    const minX = Math.min(...pointsX)
    const minIdx = pointsX.findIndex(x => x === minX)
    const minXRad = startAngle + minIdx * anglePerPoint
    const res = [{startAngle, endAngle: minXRad, edgeCount: minIdx}, {startAngle: minXRad, endAngle, edgeCount: edgeCount - minIdx}]
    console.log('splitAngle--res', res)
    return res
  } else if (endAngle <= PI) {
    return [{startAngle, endAngle}, {startAngle: endAngle, endAngle: endAngle}]
  } else {
    return [{startAngle: startAngle, endAngle: startAngle}, {startAngle, endAngle}]
  }
}

/**************************************************************************************
 * tools
 **************************************************************************************/

function getOption(startAngleOrOption, endAngle, rx = 100, ry = 80, h = 20, ir = .8, color = '#55a8cf') {
  return typeof startAngleOrOption === 'object'
    ? startAngleOrOption
    : {startAngle: startAngleOrOption, endAngle, rx, ry, h, ir, color}
}

